/*************************************************
* Common ASN.1 Objects Header File               *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#ifndef BOTAN_ASN1_OBJ_H__
#define BOTAN_ASN1_OBJ_H__

#include <botan/asn1.h>
#include <vector>
#include <map>

namespace Botan {

/*************************************************
* Algorithm Identifier                           *
*************************************************/
class AlgorithmIdentifier
   {
   public:
      OID oid;
      SecureVector<byte> parameters;

      AlgorithmIdentifier() {}
      AlgorithmIdentifier(const OID&, const SecureVector<byte>&);
   };

/*************************************************
* X.509 Time                                     *
*************************************************/
class X509_Time
   {
   public:
      std::string as_string() const;
      std::string readable_string() const;
      u64bit to_seconds() const;
      bool time_is_set() const;

      ASN1_Tag tagging() const;

      s32bit cmp(const X509_Time&) const;

      X509_Time(u64bit);
      X509_Time(const std::string& = "");
      X509_Time(const std::string&, ASN1_Tag);
   private:
      bool passes_sanity_check() const;
      u32bit year, month, day, hour, minute, second;
      ASN1_Tag tag;
   };

/*************************************************
* Simple String                                  *
*************************************************/
class ASN1_String
   {
   public:
      std::string value() const;
      ASN1_Tag tagging() const;

      ASN1_String(const std::string& = "");
      ASN1_String(const std::string&, ASN1_Tag);
   private:
      static const byte VALID_CODEPOINTS[256];
      bool valid_char(char, byte) const;
      std::string str;
      ASN1_Tag tag;
   };

/*************************************************
* Distinguished Name                             *
*************************************************/
class X509_DN
   {
   public:
      std::string common_name() const;
      std::string country_code() const;
      std::string organization() const;
      std::string organizational_unit() const;
      std::string locality() const;
      std::string state() const;
      std::string email() const;

      void add_attribute(const std::string&, const ASN1_String&);
      void add_attribute(const OID&, const ASN1_String&);

      std::map<OID, ASN1_String> get_attributes() const;

      X509_DN();
      X509_DN(const std::map<OID, ASN1_String>&);
      X509_DN(const std::map<std::string, std::string>&);
   private:
      std::string do_lookup(const OID&) const;
      std::map<OID, ASN1_String> dn_info;
   };

/*************************************************
* Alternative Name                               *
*************************************************/
class AlternativeName
   {
   public:
      std::string uri() const;
      std::string dns_name() const;
      std::string email() const;
      bool has_items() const;

      AlternativeName(const std::string& = "", const std::string& = "",
                      const std::string& = "");
   private:
      std::string URI, DNS, RFC822;
   };

/*************************************************
* Comparison Operations                          *
*************************************************/
bool operator==(const AlgorithmIdentifier&, const AlgorithmIdentifier&);
bool operator!=(const AlgorithmIdentifier&, const AlgorithmIdentifier&);

bool operator==(const X509_Time&, const X509_Time&);
bool operator!=(const X509_Time&, const X509_Time&);
bool operator<=(const X509_Time&, const X509_Time&);
bool operator>=(const X509_Time&, const X509_Time&);

bool operator==(const ASN1_String&, const ASN1_String&);
bool operator!=(const ASN1_String&, const ASN1_String&);

bool operator==(const X509_DN&, const X509_DN&);
bool operator!=(const X509_DN&, const X509_DN&);

/*************************************************
* DER Encoding Functions                         *
*************************************************/
namespace DER {

void encode(DER_Encoder&, const AlgorithmIdentifier&);
void encode(DER_Encoder&, const X509_Time&);
void encode(DER_Encoder&, const ASN1_String&);
void encode(DER_Encoder&, const ASN1_String&,
            ASN1_Tag, ASN1_Tag = CONTEXT_SPECIFIC);
void encode(DER_Encoder&, const X509_DN&);
void encode(DER_Encoder&, const AlternativeName&);
void encode(DER_Encoder&, Key_Constraints);

}

/*************************************************
* BER Decoding Functions                         *
*************************************************/
namespace BER {

void decode(BER_Decoder&, AlgorithmIdentifier&);
void decode(BER_Decoder&, X509_Time&);
void decode(BER_Decoder&, ASN1_String&);
void decode(BER_Decoder&, ASN1_String&, ASN1_Tag, ASN1_Tag);
void decode(BER_Decoder&, X509_DN&);
void decode(BER_Decoder&, AlternativeName&);
void decode(BER_Decoder&, Key_Constraints&);

}

}

#endif
