/*************************************************
* X.509 Certificates Header File                 *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#ifndef BOTAN_X509_CERTS_H__
#define BOTAN_X509_CERTS_H__

#include <botan/x509.h>
#include <botan/asn1_obj.h>
#include <map>

namespace Botan {

/*************************************************
* X.509 Certificate                              *
*************************************************/
class X509_Certificate
   {
   public:
      struct X509_Error : public Exception
         {
         X509_Error(const std::string& error) :
            Exception("X509_Certificate: " + error) {}
         };

      u32bit x509_version() const;
      bool valid_time(u64bit) const;
      X509_PublicKey* subject_public_key() const;

      std::string subject_info(const std::string&) const;
      std::string issuer_info(const std::string&) const;

      AlgorithmIdentifier signature_algorithm() const;
      SecureVector<byte> tbs_certificate() const;
      SecureVector<byte> signature() const;
      bool self_signed() const;

      bool is_ca_cert() const;
      u32bit path_len_constraint() const;
      Key_Constraints key_constraints() const;
      bool ext_key_constraints(const OID&) const;
      SecureVector<byte> authority_key_id() const;
      SecureVector<byte> subject_key_id() const;

      bool operator==(const X509_Certificate&) const;

      void encode(Pipe&, X509_Encoding = PEM) const;
      std::string PEM_encode() const;

      X509_Certificate(DataSource&, X509_Encoding = PEM);
   private:
      friend class X509_CA;
      friend class X509_Store;
      X509_Certificate() {}
      void decode_cert(DataSource&);
      void handle_v3_extension(const OID&, bool, const SecureVector<byte>&);
      std::map<std::string, std::string> subject, subject_alt;
      std::map<std::string, std::string> issuer, issuer_alt;
      SecureVector<byte> v3_issuer_key_id, v3_subject_key_id;
      SecureVector<byte> v2_issuer_key_id, v2_subject_key_id;
      SecureVector<byte> tbs_cert_data, pub_key, sig, serial;
      X509_Time start, end;
      AlgorithmIdentifier sig_algo;
      Key_Constraints constraints;
      std::vector<OID> ex_constraints;
      u32bit version, max_path_len;
      bool is_ca;
      mutable bool trusted;
   };

/*************************************************
* X.509 Certificate Comparison                   *
*************************************************/
bool operator!=(const X509_Certificate&, const X509_Certificate&);

}

#endif
