/*************************************************
* AlternativeName Source File                    *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/asn1_obj.h>

namespace Botan {

/*************************************************
* Create an AlternativeName                      *
*************************************************/
AlternativeName::AlternativeName(const std::string& email_addr,
                                 const std::string& uri,
                                 const std::string& dns)
   {
   RFC822 = email_addr;
   URI = uri;
   DNS = dns;
   }

/*************************************************
* Return if this object has anything useful      *
*************************************************/
bool AlternativeName::has_items() const
   {
   return ((RFC822 != "") || (DNS != "") || (URI != ""));
   }

/*************************************************
* Return the stored URI, if available            *
*************************************************/
std::string AlternativeName::uri() const
   {
   return URI;
   }

/*************************************************
* Return the stored DNS name, if available       *
*************************************************/
std::string AlternativeName::dns_name() const
   {
   return DNS;
   }

/*************************************************
* Return the stored email address, if available  *
*************************************************/
std::string AlternativeName::email() const
   {
   return RFC822;
   }

namespace DER {

/*************************************************
* DER encode a AlternativeName                   *
*************************************************/
void encode(DER_Encoder& encoder, const AlternativeName& alt_name)
   {
   encoder.start_sequence();
   if(alt_name.email() != "")
      encoder.add_object(ASN1_Tag(1), CONTEXT_SPECIFIC, alt_name.email());
   if(alt_name.dns_name() != "")
      encoder.add_object(ASN1_Tag(2), CONTEXT_SPECIFIC, alt_name.dns_name());
   if(alt_name.uri() != "")
      encoder.add_object(ASN1_Tag(6), CONTEXT_SPECIFIC, alt_name.uri());
   encoder.end_sequence();
   }

}

namespace BER {

/*************************************************
* Decode a BER encoded AlternativeName           *
*************************************************/
void decode(BER_Decoder& source, AlternativeName& alt_name)
   {
   BER_Decoder names = BER::get_subsequence(source);

   std::string email_addr, uri, dns;
   while(names.more_items())
      {
      const ASN1_Tag email_tag = ASN1_Tag(1 | CONTEXT_SPECIFIC);
      const ASN1_Tag dns_tag = ASN1_Tag(2 | CONTEXT_SPECIFIC);
      const ASN1_Tag uri_tag = ASN1_Tag(6 | CONTEXT_SPECIFIC);

      BER_Object name_obj = names.get_next_object();

      if(name_obj.tag == email_tag)
         email_addr = BER::to_string(name_obj);
      else if(name_obj.tag == dns_tag)
         dns = BER::to_string(name_obj);
      else if(name_obj.tag == uri_tag)
         uri = BER::to_string(name_obj);
      }

   alt_name = AlternativeName(email_addr, uri, dns);
   }

}

}
