/*************************************************
* Simple ASN.1 String Types Source File          *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/asn1_obj.h>

namespace Botan {

/*************************************************
* Create an ASN1_String                          *
*************************************************/
ASN1_String::ASN1_String(const std::string& str_arg, ASN1_Tag t) : tag(t)
   {
   if(tag != NUMERIC_STRING && tag != PRINTABLE_STRING &&
      tag != VISIBLE_STRING && tag != IA5_STRING)
      throw Invalid_Argument("ASN1_String: Invalid tag " + to_string(tag));

   byte mask = 0;
   if(tag == NUMERIC_STRING)
      mask = 0x01;
   else if(tag == PRINTABLE_STRING)
      mask = 0x02;
   else if(tag == VISIBLE_STRING)
      mask = 0x04;
   else if(tag == IA5_STRING)
      mask = 0x08;

   for(u32bit j = 0; j != str_arg.size(); j++)
      if(!valid_char(str_arg[j], mask))
         throw Invalid_Argument("ASN1_String: String contains invalid chars");

   str = str_arg;
   }

/*************************************************
* Create an ASN1_String                          *
*************************************************/
ASN1_String::ASN1_String(const std::string& str_arg)
   {
   byte mask = 0x01;

   for(u32bit j = 0; j != str_arg.size(); j++)
      {
      if(str_arg[j] == '@') mask = 0x08;
      while(!valid_char(str_arg[j], mask) && mask <= 0x80)
         mask <<= 1;
      }

   if(mask == 0x01)      tag = NUMERIC_STRING;
   else if(mask == 0x02) tag = PRINTABLE_STRING;
   else if(mask == 0x04) tag = VISIBLE_STRING;
   else if(mask == 0x08) tag = IA5_STRING;
   else
      throw Encoding_Error("No way to encode " + str_arg +
                           " as an ASN.1 string");

   str = str_arg;
   }

/*************************************************
* Check the validity of input characters         *
*************************************************/
bool ASN1_String::valid_char(char c, byte mask) const
   {
   byte point = (byte)c;
   return (VALID_CODEPOINTS[point] & mask);
   }

/*************************************************
* Return the value of the string                 *
*************************************************/
std::string ASN1_String::value() const
   {
   return str;
   }

/*************************************************
* Return the type of this string object          *
*************************************************/
ASN1_Tag ASN1_String::tagging() const
   {
   return tag;
   }

/*************************************************
* Compare two ASN1_Strings for equality          *
*************************************************/
bool operator==(const ASN1_String& s1, const ASN1_String& s2)
   {
   return (s1.value() == s2.value());
   }

/*************************************************
* Compare two ASN1_Strings for inequality        *
*************************************************/
bool operator!=(const ASN1_String& s1, const ASN1_String& s2)
   {
   return (s1.value() != s2.value());
   }

namespace DER {

/*************************************************
* DER encode an ASN1_String                      *
*************************************************/
void encode(DER_Encoder& encoder, const ASN1_String& string)
   {
   encoder.add_object(string.tagging(), UNIVERSAL, string.value());
   }

/*************************************************
* DER encode an ASN1_String, IMPLICIT tag        *
*************************************************/
void encode(DER_Encoder& encoder, const ASN1_String& string,
            ASN1_Tag type_tag, ASN1_Tag class_tag)
   {
   encoder.add_object(type_tag, class_tag, string.value());
   }

}

namespace BER {

/*************************************************
* Decode a BER encoded ASN1_String               *
*************************************************/
void decode(BER_Decoder& source, ASN1_String& string)
   {
   BER_Object ber_obj = source.get_next_object();
   string = ASN1_String(BER::to_string(ber_obj), ber_obj.tag);
   }

/*************************************************
* Decode a BER encoded ASN1_String               *
*************************************************/
void decode(BER_Decoder& source, ASN1_String& string,
            ASN1_Tag expected_tag, ASN1_Tag real_tag)
   {
   BER_Object ber_obj = source.get_next_object();
   if(ber_obj.tag != expected_tag)
      throw Decoding_Error("Unexpected tag in BER string decoding");
   string = ASN1_String(BER::to_string(ber_obj), real_tag);
   }

}

}
