/*************************************************
* OID Registry Source File                       *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/oids.h>
#include <botan/exceptn.h>
#include <map>

namespace Botan {

namespace OIDS {

namespace {

std::map<OID, std::string> oid_to_str;
std::map<std::string, OID> str_to_oid;

}

/*************************************************
* Register an OID to string mapping              *
*************************************************/
void add_oid(const OID& oid, const std::string& name)
   {
   oid_to_str[oid] = name;
   str_to_oid[name] = oid;
   }

/*************************************************
* Do an OID to string lookup                     *
*************************************************/
std::string lookup(const OID& oid)
   {
   std::map<OID, std::string>::const_iterator info = oid_to_str.find(oid);
   if(info == oid_to_str.end())
      return "";
   return info->second;
   }

/*************************************************
* Do a string to OID lookup                      *
*************************************************/
OID lookup(const std::string& name)
   {
   std::map<std::string, OID>::const_iterator info = str_to_oid.find(name);
   if(info == str_to_oid.end())
      throw Lookup_Error("No known OID for " + name);
   return info->second;
   }

}

namespace {

/*************************************************
* OID loading helper function                    *
*************************************************/
void add_oid(const std::string& oid_str, const std::string& name)
   {
   OID oid(oid_str);
   OIDS::add_oid(oid, name);
   }

/*************************************************
* Register a one-way string to OID mapping       *
*************************************************/
void add_oid_alias(const std::string alias, const std::string& oid_str)
   {
   OID oid(oid_str);
   OIDS::str_to_oid[alias] = oid;
   }

}

/*************************************************
* Load all of the default OIDs                   *
*************************************************/
void setup_oids()
   {
   add_oid("1.2.840.113549.1.1.1", "RSA");
   add_oid("1.2.840.10040.4.1", "DSA");
   add_oid("1.2.840.10046.2.1", "DH");

   add_oid("1.3.14.3.2.7", "DES/CBC");
   add_oid("1.2.840.113549.3.7", "TripleDES/CBC");

   add_oid("1.2.840.10040.4.3", "DSA/EMSA1(SHA-1)/DER");
   add_oid("1.2.840.113549.1.1.2", "RSA/EMSA3(MD2)");
   add_oid("1.2.840.113549.1.1.4", "RSA/EMSA3(MD5)");
   add_oid("1.2.840.113549.1.1.5", "RSA/EMSA3(SHA-1)");

   add_oid("1.2.840.113549.1.5.13", "PKCS8.PBES2");
   add_oid("1.2.840.113549.1.5.12", "PKCS8.PBKDF2");

   add_oid("1.2.840.113549.1.9.1", "PKCS9.EmailAddress");

   add_oid("2.5.4.3", "X500.CommonName");
   add_oid("2.5.4.6", "X500.Country");
   add_oid("2.5.4.8", "X500.State");
   add_oid("2.5.4.7", "X500.Locality");
   add_oid("2.5.4.10", "X500.Organization");
   add_oid("2.5.4.11", "X500.OrganizationalUnit");

   add_oid("2.5.29.15", "X509v3.KeyUsage");
   add_oid("2.5.29.19", "X509v3.BasicConstraints");
   add_oid("2.5.29.37", "X509v3.ExtendedKeyUsage");
   add_oid("2.5.29.35", "X509v3.AuthorityKeyIdentifier");
   add_oid("2.5.29.14", "X509v3.SubjectKeyIdentifier");
   add_oid("2.5.29.17", "X509v3.SubjectAlternativeName");
   add_oid("2.5.29.18", "X509v3.IssuerAlternativeName");

   add_oid("1.3.6.1.5.5.7.3.1", "PKIX.ServerAuth");
   add_oid("1.3.6.1.5.5.7.3.2", "PKIX.ClientAuth");
   add_oid("1.3.6.1.5.5.7.3.3", "PKIX.CodeSigning");
   add_oid("1.3.6.1.5.5.7.3.4", "PKIX.EmailProtection");
   add_oid("1.3.6.1.5.5.7.3.8", "PKIX.TimeStamping");

   add_oid_alias("DSA/EMSA1(SHA-1)", "1.2.840.10040.4.3");
   }

}
