/*************************************************
* RSA Source File                                *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/rsa.h>
#include <botan/numthry.h>
#include <botan/keypair.h>

namespace Botan {

/*************************************************
* RSA_PublicKey Constructor                      *
*************************************************/
RSA_PublicKey::RSA_PublicKey(const BigInt& mod, const BigInt& exp)
   {
   if_initialize(mod, exp, true);
   if(!check_params())
      throw Invalid_Argument("RSA: invalid parameters for public key");
   }

/*************************************************
* Decode a X.509 RSA public key                  *
*************************************************/
RSA_PublicKey::RSA_PublicKey(DataSource& source, X509_Encoding encoding)
   {
   X509::decode(*this, source, encoding);
   }

/*************************************************
* Return the name of this algorithm              *
*************************************************/
std::string RSA_PublicKey::algo_name() const
   {
   return "RSA";
   }

/*************************************************
* RSA Encryption Function                        *
*************************************************/
SecureVector<byte> RSA_PublicKey::encrypt(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode_1363(public_op(temp), n.bytes());
   }

/*************************************************
* RSA Verification Function                      *
*************************************************/
SecureVector<byte> RSA_PublicKey::verify(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode(public_op(temp));
   }

/*************************************************
* RSA Public Operation                           *
*************************************************/
BigInt RSA_PublicKey::public_op(const BigInt& i) const
   {
   if(i >= n || i.is_negative())
      throw Invalid_Argument("RSA::public_op: i >= n || i < 0");
   return powermod_e_n(i);
   }

/*************************************************
* RSA_PrivateKey Constructor                     *
*************************************************/
RSA_PrivateKey::RSA_PrivateKey(const BigInt& prime1, const BigInt& prime2,
                               const BigInt& exp, const BigInt& d_exp,
                               const BigInt& modulus) :
   RSA_PublicKey(((!modulus) ? prime1 * prime2 : modulus), exp)
   {
   if(prime1 < 3 || prime2 < 3)
      throw Invalid_Argument("RSA_PrivateKey: invalid prime(s)");
   if(d_exp != 0 && d_exp < 3)
      throw Invalid_Argument("RSA_PrivateKey: invalid decryption exponent");

   p = prime1;
   q = prime2;
   d = (!d_exp) ? inverse_mod(e, lcm(p - 1, q - 1)) : d_exp;
   precompute();
   }

/*************************************************
* Create a RSA private key                       *
*************************************************/
RSA_PrivateKey::RSA_PrivateKey(u32bit bits, const BigInt& exp)
   {
   if(bits < 512)
      throw Invalid_Argument("RSA: Can't make a key that is only " +
                             to_string(bits) + " bits long");
   if(exp < 3 || exp % 2 == 0)
      throw Invalid_Argument("RSA: Invalid encryption exponent");

   e = exp;
   p = random_prime((bits + 1) / 2, LongTermKey, e);
   q = random_prime(bits - p.bits(), LongTermKey, e);
   n = p * q;
   d = inverse_mod(e, lcm(p - 1, q - 1));
   precompute();

   KeyPair::check_key(get_pk_encryptor(*this, "EME1(SHA-1)"),
                      get_pk_decryptor(*this, "EME1(SHA-1)")
      );

   KeyPair::check_key(get_pk_signer(*this, "EMSA4(SHA-1)"),
                      get_pk_verifier(*this, "EMSA4(SHA-1)")
      );
   }

/*************************************************
* Decrypt/decode a PKCS #8 RSA private key       *
*************************************************/
RSA_PrivateKey::RSA_PrivateKey(DataSource& source, const std::string& pass,
                               X509_Encoding encoding)
   {
   PKCS8::decrypt_key(*this, source, pass, encoding);
   }

/*************************************************
* RSA Decryption Operation                       *
*************************************************/
SecureVector<byte> RSA_PrivateKey::decrypt(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode(if_private_op(temp));
   }

/*************************************************
* RSA Signature Operation                        *
*************************************************/
SecureVector<byte> RSA_PrivateKey::sign(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode_1363(if_private_op(temp), n.bytes());
   }

/*************************************************
* Check Private RSA Parameters                   *
*************************************************/
bool RSA_PrivateKey::check_params() const
   {
   if(!IF_Scheme_PrivateKey::check_params())
      return false;
   if((e * d) % lcm(p - 1, q - 1) != 1)
      return false;
   return true;
   }

}
