/*************************************************
* Rabin-Williams Source File                     *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/rw.h>
#include <botan/numthry.h>
#include <botan/keypair.h>

namespace Botan {

/*************************************************
* RW_PublicKey Constructor                       *
*************************************************/
RW_PublicKey::RW_PublicKey(const BigInt& mod, const BigInt& exp)
   {
   if_initialize(mod, exp, false);
   if(e < 2 || e % 2 == 1)
      throw Invalid_Argument("RW: Invalid public exponent");
   if(n < 21 || n % 2 == 0)
      throw Invalid_Argument("RW: Invalid modulus");
   }

/*************************************************
* Decode a X.509 RW public key                   *
*************************************************/
RW_PublicKey::RW_PublicKey(DataSource& source, X509_Encoding encoding)
   {
   X509::decode(*this, source, encoding);
   }

/*************************************************
* Return the name of this algorithm              *
*************************************************/
std::string RW_PublicKey::algo_name() const
   {
   return "RW";
   }

/*************************************************
* Rabin-Williams Verification Function           *
*************************************************/
SecureVector<byte> RW_PublicKey::verify(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode(public_op(temp));
   }

/*************************************************
* Rabin-Williams Public Operation                *
*************************************************/
BigInt RW_PublicKey::public_op(const BigInt& i) const
   {
   if(i > n / 2 || i.is_negative())
      throw Invalid_Argument("Rabin-Williams::public_op: i > n / 2 || i < 0");

   BigInt t1 = powermod_e_n(i);
   if(t1 % 16 == 12) return t1;
   if(t1 % 8 == 6)   return 2*t1;

   BigInt t2 = n - t1;
   if(t2 % 16 == 12) return t2;
   if(t2 % 8 == 6)   return 2*t2;

   throw Invalid_Argument("Rabin-Williams::public_op: Invalid input");
   }

/*************************************************
* RW_PrivateKey Constructor                      *
*************************************************/
RW_PrivateKey::RW_PrivateKey(const BigInt& prime1, const BigInt& prime2,
                             const BigInt& exp, const BigInt& d_exp,
                             const BigInt& modulus)
   {
   if(exp.is_zero() || exp.is_odd())
      throw Invalid_Argument("RW_PrivateKey: Exponent is invalid");

   e = exp;
   p = prime1;
   q = prime2;
   n = (!modulus) ? p * q : modulus;
   d = (!d_exp) ? inverse_mod(e, lcm(p - 1, q - 1) / 2) : d_exp;
   precompute();
   }

/*************************************************
* Create a Rabin-Williams private key            *
*************************************************/
RW_PrivateKey::RW_PrivateKey(u32bit bits, const BigInt& exp)
   {
   if(bits < 512)
      throw Invalid_Argument("Rabin-Williams: Can't make a key that is only " +
                             to_string(bits) + " bits long");
   if(exp < 2 || exp.is_odd())
      throw Invalid_Argument("Rabin-Williams: Invalid exponent");

   e = exp;

   p = random_prime((bits + 1) / 2, LongTermKey, e / 2, 3, 4);
   q = random_prime(bits - p.bits(), LongTermKey, e / 2,
                    ((p % 8 == 3) ? 7 : 3), 8);

   n = p * q;
   d = inverse_mod(e, lcm(p - 1, q - 1) / 2);
   precompute();

   KeyPair::check_key(get_pk_signer(*this, "EMSA2(SHA-1)"),
                      get_pk_verifier(*this, "EMSA2(SHA-1)")
      );
   }

/*************************************************
* Decrypt/decode a PKCS #8 RW private key        *
*************************************************/
RW_PrivateKey::RW_PrivateKey(DataSource& source, const std::string& pass,
                             X509_Encoding encoding)
   {
   PKCS8::decrypt_key(*this, source, pass, encoding);
   }

/*************************************************
* Rabin-Williams Signature Operation             *
*************************************************/
SecureVector<byte> RW_PrivateKey::sign(const byte in[], u32bit len) const
   {
   BigInt temp(in, len);
   return encode_1363(private_op(temp), n.bytes());
   }

/*************************************************
* Rabin-Williams Private Key Operation           *
*************************************************/
BigInt RW_PrivateKey::private_op(const BigInt& i) const
   {
   if(i >= n || i.is_negative())
      throw Invalid_Argument("Rabin-Williams::private_op: i >= n || i < 0");
   if(i % 16 != 12)
      throw Invalid_Argument("Rabin-Williams::public_op: i % 16 != 12");

   BigInt t;
   if(jacobi(i, n) == 1)
      t = if_private_op(i);
   else
      t = if_private_op(i / 2);
   return std::min(t, n - t);
   }

/*************************************************
* Check Private Rabin-Williams Parameters        *
*************************************************/
bool RW_PrivateKey::check_params() const
   {
   if(!IF_Scheme_PrivateKey::check_params())
      return false;
   if((e * d) % (lcm(p - 1, q - 1) / 2) != 1)
      return false;
   return true;
   }

}
