/*************************************************
* X.509 Certificate Authority Source File        *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/x509_ca.h>
#include <botan/lookup.h>
#include <botan/look_pk.h>
#include <botan/oids.h>
#include <botan/rng.h>
#include <typeinfo>
#include <memory>

namespace Botan {

/*************************************************
* Load the keys and public certificate           *
*************************************************/
X509_CA::X509_CA(const X509_Certificate& c,
                 const PKCS8_PrivateKey& key) : cert(c)
   {
   if(!cert.is_ca_cert())
      throw Invalid_Argument("X509_CA: This certificate is not for a CA");

   std::string padding;
   Signature_Format format;
   if(key.algo_name() == "RSA")
      { padding = "EMSA3(SHA-1)"; format = IEEE_1363; }
   else if(key.algo_name() == "DSA")
      { padding = "EMSA1(SHA-1)"; format = DER_SEQUENCE; }
   else
      throw Invalid_Argument("Unknown X.509 signing key type: " +
                             key.algo_name());

   ca_sig_algo.oid = OIDS::lookup(key.algo_name() + "/" + padding);
   ca_sig_algo.parameters = key.DER_encode_params();

   try {
      const PK_Signing_Key& sig_key = dynamic_cast<const PK_Signing_Key&>(key);
      signer = get_pk_signer(sig_key, padding, format);
   }
   catch(std::bad_cast)
      {
      throw Decoding_Error("X509_CA: " + key.algo_name() + " cannot sign");
      }

   ca_dn_info = cert.subject;
   }

/*************************************************
* Create a new CA certificate                    *
*************************************************/
X509_CA::X509_CA(const PKCS8_PrivateKey& key,
                 X509_Cert_Options& opts)
   {
   std::string padding;
   Signature_Format format;
   if(key.algo_name() == "RSA")
      { padding = "EMSA3(SHA-1)"; format = IEEE_1363; }
   else if(key.algo_name() == "DSA")
      { padding = "EMSA1(SHA-1)"; format = DER_SEQUENCE; }
   else
      throw Invalid_Argument("Unknown X.509 signing key type: " +
                             key.algo_name());

   opts.CA_key();
   opts.self_signed = true;
   opts.set_constraints(Key_Constraints(KEY_CERT_SIGN | CRL_SIGN));

   ca_sig_algo.oid = OIDS::lookup(key.algo_name() + "/" + padding);
   ca_sig_algo.parameters = key.DER_encode_params();

   try {
      const PK_Signing_Key& sig_key = dynamic_cast<const PK_Signing_Key&>(key);
      signer = get_pk_signer(sig_key, padding, format);
      cert = sign_key(key, opts);
   }
   catch(std::bad_cast)
      {
      throw Invalid_Argument("X509_CA: The private key provided is not for "
                             "a signature algorithm");
      }

   ca_dn_info = cert.subject;
   }

/*************************************************
* Find the allowable key constraints             *
*************************************************/
Key_Constraints X509_CA::find_constraints(const X509_PublicKey& pub_key,
                                          Key_Constraints other) const
   {
   const X509_PublicKey* key = &pub_key;
   u32bit constraints = 0;

   if(dynamic_cast<const PK_Encrypting_Key*>(key))
      constraints |= KEY_ENCIPHERMENT | DATA_ENCIPHERMENT;

   if(dynamic_cast<const PK_Key_Agreement_Key*>(key))
      constraints |= KEY_AGREEMENT;

   if(dynamic_cast<const PK_Verifying_wo_MR_Key*>(key) ||
      dynamic_cast<const PK_Verifying_with_MR_Key*>(key))
      constraints |= DIGITAL_SIGNATURE | NON_REPUDATION;

   if(other)
      constraints &= other;

   return Key_Constraints(constraints);
   }

/*************************************************
* Create a new certificate                       *
*************************************************/
X509_Certificate X509_CA::sign_key(const X509_PublicKey& pub_key,
                                   const X509_Cert_Options& opts) const
   {
   Pipe key_encoder;
   key_encoder.start_msg();
   X509::encode(pub_key, key_encoder, RAW_BER);
   key_encoder.end_msg();

   Key_Constraints constraints;

   if(opts.is_ca)
      {
      if(!dynamic_cast<const PK_Signing_Key*>(&pub_key))
         throw Invalid_Argument("X509_CA: The private key provided is not for "
                                "a signature algorithm");
      constraints = Key_Constraints(KEY_CERT_SIGN | CRL_SIGN);
      }
   else
      constraints = find_constraints(pub_key, opts.constraints);

   return make_cert(key_encoder.read_all(), opts, constraints);
   }

/*************************************************
* Create a new certificate                       *
*************************************************/
X509_Certificate X509_CA::make_cert(const SecureVector<byte>& pub_key,
                                    const X509_Cert_Options& opts,
                                    Key_Constraints constraints) const
   {
   X509_Certificate new_cert;

   opts.sanity_check();

   new_cert.version = 2;
   new_cert.is_ca = opts.is_ca;
   new_cert.max_path_len = 0;
   new_cert.constraints = constraints;
   new_cert.ex_constraints = opts.ex_constraints;
   new_cert.start = opts.start;
   new_cert.end = opts.end;
   new_cert.sig_algo = ca_sig_algo;
   new_cert.v3_issuer_key_id = cert.subject_key_id();

   new_cert.subject["X500.CommonName"] = opts.name;
   new_cert.subject["X500.Country"] = opts.country;
   new_cert.subject["X500.State"] = opts.state;
   new_cert.subject["X500.Locality"] = opts.locality;
   new_cert.subject["X500.Organization"] = opts.org;
   new_cert.subject["X500.OrganizationalUnit"] = opts.org_unit;
   new_cert.subject_alt["RFC822"] = opts.email;
   new_cert.subject_alt["URI"] = opts.uri;
   new_cert.subject_alt["DNS"] = opts.dns;

   if(opts.self_signed)
      new_cert.issuer = new_cert.subject;
   else
      new_cert.issuer = ca_dn_info;

   new_cert.serial.create(10);
   Global_RNG::randomize(new_cert.serial, new_cert.serial.size());

   new_cert.pub_key = pub_key;

   std::auto_ptr<HashFunction> hash(get_hash("SHA-1"));
   hash->update(new_cert.pub_key, new_cert.pub_key.size());
   new_cert.v3_subject_key_id = hash->final();

   DER_Encoder tbs_cert;

   tbs_cert.start_explicit(ASN1_Tag(0));
   DER::encode(tbs_cert, new_cert.version);
   tbs_cert.end_explicit(ASN1_Tag(0));

   DER::encode(tbs_cert, BigInt(new_cert.serial, new_cert.serial.size()));
   DER::encode(tbs_cert, new_cert.sig_algo);
   DER::encode(tbs_cert, X509_DN(new_cert.issuer));
   tbs_cert.start_sequence();
   DER::encode(tbs_cert, new_cert.start);
   DER::encode(tbs_cert, new_cert.end);
   tbs_cert.end_sequence();
   DER::encode(tbs_cert, X509_DN(new_cert.subject));
   tbs_cert.add_raw_octets(new_cert.pub_key);

   tbs_cert.start_explicit(ASN1_Tag(3));
   tbs_cert.start_sequence();

   DER_Encoder v3_ext;

   if(opts.is_ca)
      {
      if(opts.self_signed)
         new_cert.max_path_len = opts.path_limit;
      else
         {
         if(cert.path_len_constraint() == 0)
            throw Encoding_Error("X509_CA: This CA cannot sign new CA keys");

         new_cert.max_path_len = std::min(opts.path_limit,
                                          cert.path_len_constraint() - 1);
         }

      v3_ext.start_sequence();
      DER::encode(v3_ext, true);
      DER::encode(v3_ext, new_cert.max_path_len);
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.BasicConstraints", true);
      }

   if(new_cert.subject_key_id().size())
      {
      DER::encode(v3_ext, new_cert.subject_key_id(), OCTET_STRING);
      do_ext(tbs_cert, v3_ext, "X509v3.SubjectKeyIdentifier", false);
      }

   if(new_cert.authority_key_id().size())
      {
      v3_ext.start_sequence();
      v3_ext.start_explicit(ASN1_Tag(0));
      DER::encode(v3_ext, new_cert.authority_key_id(), OCTET_STRING);
      v3_ext.end_explicit(ASN1_Tag(0));
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.AuthorityKeyIdentifier", false);
      }

   AlternativeName subject_alt(opts.email, opts.uri, opts.dns);
   if(subject_alt.has_items())
      {
      DER::encode(v3_ext, subject_alt);
      do_ext(tbs_cert, v3_ext, "X509v3.SubjectAlternativeName", false);
      }

   if(new_cert.constraints)
      {
      DER::encode(v3_ext, new_cert.constraints);
      do_ext(tbs_cert, v3_ext, "X509v3.KeyUsage", true);
      }

   if(new_cert.ex_constraints.size())
      {
      v3_ext.start_sequence();
      for(u32bit j = 0; j != new_cert.ex_constraints.size(); j++)
         DER::encode(v3_ext, new_cert.ex_constraints[j]);
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.ExtendedKeyUsage", false);
      }

   tbs_cert.end_sequence();
   tbs_cert.end_explicit(ASN1_Tag(3));

   new_cert.tbs_cert_data = tbs_cert.get_contents();

   new_cert.sig = signer->sign_message(new_cert.tbs_certificate());

   return new_cert;
   }

/*************************************************
* Handle encoding a v3 extension                 *
*************************************************/
void X509_CA::do_ext(DER_Encoder& new_cert, DER_Encoder& extension,
                     const std::string& oid, bool critical) const
   {
   SecureVector<byte> ext_val = extension.get_contents();

   new_cert.start_sequence();
   DER::encode(new_cert, OIDS::lookup(oid));
   if(critical)
      DER::encode(new_cert, critical);
   DER::encode(new_cert, ext_val, OCTET_STRING);
   new_cert.end_sequence();
   }

/*************************************************
* Return the CA's certificate                    *
*************************************************/
X509_Certificate X509_CA::ca_certificate() const
   {
   return cert;
   }

/*************************************************
* X509_CA Destructor                             *
*************************************************/
X509_CA::~X509_CA()
   {
   delete signer;
   }

}
