/*************************************************
* X.509 Certificates Source File                 *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/x509cert.h>
#include <botan/bigint.h>
#include <botan/oids.h>
#include <botan/pem.h>

namespace Botan {

namespace {

/*************************************************
* Deref aliases in a subject/issuer info request *
*************************************************/
std::string deref_info_field(const std::string& info)
   {
   if(info == "Name" || info == "CommonName" || info == "CN")
      return "X500.CommonName";
   if(info == "Country" || info == "C")
      return "X500.Country";
   if(info == "Organization" || info == "O")
      return "X500.Organization";
   if(info == "Organization Unit" || info == "OU")
      return "X500.OrganizationalUnit";
   if(info == "Locality" || info == "L")
      return "X500.Locality";
   if(info == "State" || info == "Province" || info == "ST" || info == "PR")
      return "X500.State";
   if(info == "Email")
      return "RFC822";
   if(info == "URL")
      return "URI";
   return info;
   }

/*************************************************
* Get information from the DistinguishedName     *
*************************************************/
void load_info(std::map<std::string, std::string>& name,
               std::map<std::string, std::string>& alt_name,
               const X509_DN& dn_info)
   {
   if(dn_info.common_name() != "")
      name["X500.CommonName"] = dn_info.common_name();
   if(dn_info.country_code() != "")
      name["X500.Country"] = dn_info.country_code();
   if(dn_info.organization() != "")
      name["X500.Organization"] = dn_info.organization();
   if(dn_info.organizational_unit() != "")
      name["X500.Organizational Unit"] = dn_info.organizational_unit();
   if(dn_info.locality() != "")
      name["X500.Locality"] = dn_info.locality();
   if(dn_info.state() != "")
      name["X500.State"] = dn_info.state();
   if(dn_info.email() != "")
      alt_name["RFC822"] = dn_info.email();
   }

/*************************************************
* Get information from the alternative name      *
*************************************************/
void load_info(std::map<std::string, std::string>& alt_name,
               const AlternativeName& alt_info)
   {
   if(alt_info.email() != "")
      alt_name["RFC822"] = alt_info.email();
   if(alt_info.uri() != "")
      alt_name["URI"] = alt_info.uri();
   if(alt_info.dns_name() != "")
      alt_name["DNS"] = alt_info.dns_name();
   }

}

/*************************************************
* X509_Certificate Constructor                   *
*************************************************/
X509_Certificate::X509_Certificate(DataSource& in, X509_Encoding encoding)
   {
   is_ca = false;
   trusted = false;
   max_path_len = 0;

   try {
      if(encoding == RAW_BER)
         decode_cert(in);
      else
         {
         Pipe ber;
         ber.process_msg(pem_decode_check_label(in, "CERTIFICATE"));
         decode_cert(ber);
         }
      }
   catch(Decoding_Error)
      {
      throw X509_Error("Certificate decoding failed");
      }
   }

/*************************************************
* Decode a BER encoded X.509 Certificate         *
*************************************************/
void X509_Certificate::decode_cert(DataSource& source)
   {
   BER_Decoder ber(source);
   BER_Decoder sequence = BER::get_subsequence(ber);
   tbs_cert_data = BER::get_subsequence(sequence).get_remaining();

   AlgorithmIdentifier sig_algo_outer;
   BER::decode(sequence, sig_algo_outer);
   BER::decode(sequence, sig, BIT_STRING);
   sequence.verify_end();

   BER_Decoder tbs_cert(tbs_cert_data);

   BER::decode_optional(tbs_cert, version, ASN1_Tag(0),
                        ASN1_Tag(CONSTRUCTED | CONTEXT_SPECIFIC));

   if(version > 2)
      throw X509_Error("Unknown X.509 version " + to_string(version+1));
   if(version < 2) { is_ca = true; max_path_len = 0xFFFFFFFF; }

   BigInt serial_number;
   BER::decode(tbs_cert, serial_number);
   serial = Botan::encode(serial_number);

   BER::decode(tbs_cert, sig_algo);

   if(sig_algo != sig_algo_outer)
      throw X509_Error("Algorithm identifier mismatch");

   X509_DN dn_issuer;
   BER::decode(tbs_cert, dn_issuer);
   load_info(issuer, issuer_alt, dn_issuer);

   BER_Decoder validity = BER::get_subsequence(tbs_cert);
   BER::decode(validity, start);
   BER::decode(validity, end);
   validity.verify_end();

   X509_DN dn_subject;
   BER::decode(tbs_cert, dn_subject);
   load_info(subject, subject_alt, dn_subject);

   BER_Object subject_public_key = tbs_cert.get_next_object();
   if(subject_public_key.tag != ASN1_Tag(SEQUENCE | CONSTRUCTED))
      throw X509_Error("Unexpected tag for public key");
   pub_key = DER::put_in_sequence(subject_public_key.value);

   BER::decode_optional_string(tbs_cert, v2_issuer_key_id, BIT_STRING,
                               ASN1_Tag(1), CONTEXT_SPECIFIC);
   BER::decode_optional_string(tbs_cert, v2_subject_key_id, BIT_STRING,
                               ASN1_Tag(2), CONTEXT_SPECIFIC);

   BER_Object v3_exts_data = tbs_cert.get_next_object();
   if(v3_exts_data.tag == ASN1_Tag(3 | CONSTRUCTED | CONTEXT_SPECIFIC))
      {
      BER_Decoder v3_exts_decoder(v3_exts_data.value);
      BER_Decoder sequence = BER::get_subsequence(v3_exts_decoder);

      while(sequence.more_items())
         {
         OID ext_oid;
         bool ext_critical;
         SecureVector<byte> ext_value;

         BER_Decoder extension = BER::get_subsequence(sequence);
         BER::decode(extension, ext_oid);
         BER::decode_optional(extension, ext_critical,
                              BOOLEAN, UNIVERSAL, false);
         BER::decode(extension, ext_value, OCTET_STRING);
         extension.verify_end();

         handle_v3_extension(ext_oid, ext_critical, ext_value);
         }
      sequence.verify_end();
      v3_exts_decoder.verify_end();
      }
   else if(v3_exts_data.tag != NO_OBJECT)
      throw X509_Error("Unknown tag in certificate");

   if(tbs_cert.more_items())
      throw X509_Error("TBSCertificate has more items that expected");
   }

/*************************************************
* Decode a particular v3 extension               *
*************************************************/
void X509_Certificate::handle_v3_extension(const OID& oid, bool critical,
                                           const SecureVector<byte>& value)
   {
   BER_Decoder extension(value);

   if(oid == OIDS::lookup("X509v3.KeyUsage"))
      {
      BER::decode(extension, constraints);
      }
   else if(oid == OIDS::lookup("X509v3.ExtendedKeyUsage"))
      {
      BER_Decoder key_usage = BER::get_subsequence(extension);
      while(key_usage.more_items())
         {
         OID usage_oid;
         BER::decode(key_usage, usage_oid);
         ex_constraints.push_back(usage_oid);
         }
      std::sort(ex_constraints.begin(), ex_constraints.end());
      }
   else if(oid == OIDS::lookup("X509v3.BasicConstraints"))
      {
      BER_Decoder constraints = BER::get_subsequence(extension);
      BER::decode_optional(constraints, is_ca, BOOLEAN, UNIVERSAL, false);
      BER::decode_optional(constraints, max_path_len,
                           INTEGER, UNIVERSAL, 0xFFFFFFFF);
      }
   else if(oid == OIDS::lookup("X509v3.SubjectKeyIdentifier"))
      {
      BER::decode(extension, v3_subject_key_id, OCTET_STRING);
      }
   else if(oid == OIDS::lookup("X509v3.AuthorityKeyIdentifier"))
      {
      BER_Decoder key_id = BER::get_subsequence(extension);
      BER::decode_optional_string(key_id, v3_issuer_key_id, OCTET_STRING,
                                  ASN1_Tag(0), CONTEXT_SPECIFIC);
      }
   else if(oid == OIDS::lookup("X509v3.SubjectAlternativeName"))
      {
      AlternativeName alt_name;
      BER::decode(extension, alt_name);
      load_info(subject_alt, alt_name);
      }
   else
      {
      if(critical)
         throw X509_Error("Unknown critical v3 extension: " + oid.as_string());
      return;
      }

   extension.verify_end();
   }

/*************************************************
* Return a BER or PEM encoded X.509 certificate  *
*************************************************/
void X509_Certificate::encode(Pipe& pipe, X509_Encoding encoding) const
   {
   DER_Encoder encoder;

   encoder.start_sequence();
   encoder.add_raw_octets(tbs_certificate());
   DER::encode(encoder, sig_algo);
   DER::encode(encoder, sig, BIT_STRING);
   encoder.end_sequence();

   SecureVector<byte> der = encoder.get_contents();
   if(encoding == PEM)
      pipe.write(pem_encode(der, der.size(), "CERTIFICATE"));
   else
      pipe.write(der);
   }

/*************************************************
* Return a PEM encoded X.509 certificate         *
*************************************************/
std::string X509_Certificate::PEM_encode() const
   {
   Pipe pem;
   pem.start_msg();
   encode(pem, PEM);
   pem.end_msg();
   return pem.read_all_as_string();
   }

/*************************************************
* Return the X.509 version in use                *
*************************************************/
u32bit X509_Certificate::x509_version() const
   {
   return (version + 1);
   }

/*************************************************
* Check if this certificate is currently valid   *
*************************************************/
bool X509_Certificate::valid_time(u64bit current_time) const
   {
   const u32bit ALLOWABLE_SLIP = 24 * 60 * 60;

   const u64bit start_time = start.to_seconds() - ALLOWABLE_SLIP;
   const u64bit end_time = end.to_seconds() + ALLOWABLE_SLIP;

   if(current_time < start_time || current_time > end_time)
      return false;
   return true;
   }

/*************************************************
* Return information about the subject           *
*************************************************/
std::string X509_Certificate::subject_info(const std::string& info) const
   {
   std::string what = deref_info_field(info);
   std::map<std::string, std::string>::const_iterator p = subject.find(what);
   if(p != subject.end())
      return p->second;
   p = subject_alt.find(what);
   if(p != subject_alt.end())
      return p->second;
   return "";
   }

/*************************************************
* Return information about the issuer            *
*************************************************/
std::string X509_Certificate::issuer_info(const std::string& info) const
   {
   std::string what = deref_info_field(info);
   std::map<std::string, std::string>::const_iterator p = issuer.find(what);
   if(p != issuer.end())
      return p->second;
   p = issuer_alt.find(what);
   if(p != issuer_alt.end())
      return p->second;
   return "";
   }

/*************************************************
* Return the algorithm used to sign this cert    *
*************************************************/
AlgorithmIdentifier X509_Certificate::signature_algorithm() const
   {
   return sig_algo;
   }

/*************************************************
* Return the public key in this certificate      *
*************************************************/
X509_PublicKey* X509_Certificate::subject_public_key() const
   {
   DataSource_Memory source(pub_key, pub_key.size());
   return X509::load_key(source, RAW_BER);
   }

/*************************************************
* Return the TBSCertificate SEQUENCE             *
*************************************************/
SecureVector<byte> X509_Certificate::tbs_certificate() const
   {
   return DER::put_in_sequence(tbs_cert_data);
   }

/*************************************************
* Return the signature of this certificate       *
*************************************************/
SecureVector<byte> X509_Certificate::signature() const
   {
   return sig;
   }

/*************************************************
* Check if the certificate is self-signed        *
*************************************************/
bool X509_Certificate::self_signed() const
   {
   return (issuer == subject);
   }

/*************************************************
* Check if the certificate is for a CA           *
*************************************************/
bool X509_Certificate::is_ca_cert() const
   {
   if(!is_ca) return false;
   if(constraints == Key_Constraints(0))
      return true;
   return (constraints & KEY_CERT_SIGN);
   }

/*************************************************
* Return the path constaints                     *
*************************************************/
u32bit X509_Certificate::path_len_constraint() const
   {
   return max_path_len;
   }

/*************************************************
* Return the key usage constraints               *
*************************************************/
Key_Constraints X509_Certificate::key_constraints() const
   {
   return constraints;
   }

/*************************************************
* Check if the oid is in the extened key usage   *
*************************************************/
bool X509_Certificate::ext_key_constraints(const OID& oid) const
   {
   return std::binary_search(ex_constraints.begin(),
                             ex_constraints.end(), oid);
   }

/*************************************************
* Return the authority key id                    *
*************************************************/
SecureVector<byte> X509_Certificate::authority_key_id() const
   {
   return v3_issuer_key_id;
   }

/*************************************************
* Return the subject key id                      *
*************************************************/
SecureVector<byte> X509_Certificate::subject_key_id() const
   {
   return v3_subject_key_id;
   }

/*************************************************
* Compare two certificates for equality          *
*************************************************/
bool X509_Certificate::operator==(const X509_Certificate& cert) const
   {
   if(sig != cert.sig || pub_key != cert.pub_key || sig_algo != cert.sig_algo)
      return false;
   if(issuer != cert.issuer || subject != cert.subject)
      return false;
   if(serial != cert.serial || version != cert.version)
      return false;
   if(start != cert.start || end != cert.end)
      return false;
   return true;
   }

/*************************************************
* X.509 Certificate Comparison                   *
*************************************************/
bool operator!=(const X509_Certificate& cert1, const X509_Certificate& cert2)
   {
   return !(cert1 == cert2);
   }

}
