/*************************************************
* X.509 Certificate Store Source File            *
* (C) 1999-2002 The Botan Project                *
*************************************************/


#include <botan/x509stor.h>
#include <botan/pubkey.h>
#include <botan/look_pk.h>
#include <botan/oids.h>
#include <memory>

namespace Botan {

namespace {

/*************************************************
* Compare the value of unique ID fields          *
*************************************************/
bool compare_ids(const SecureVector<byte>& id1,
                 const SecureVector<byte>& id2)
   {
   if(!id1.size() || !id2.size())
      return true;
   return (id1 == id2);
   }

}

/*************************************************
* Verify a certificate's authenticity            *
*************************************************/
bool X509_Store::valid_cert(const X509_Certificate& cert) const
   {
   std::vector<u32bit> indexes;
   u32bit parent = find_parent_of(cert);
   while(true)
      {
      if(parent == NO_CERT_FOUND)
         return false;
      indexes.push_back(parent);

      if(!certs[parent].is_ca_cert())
         return false;
      if(certs[parent].trusted)
         break;
      if(certs[parent].self_signed())
         return false;
      parent = find_parent_of(certs[indexes.back()]);
      }

   const u64bit current_time = system_time();
   if(!cert.valid_time(current_time))
      return false;
   if(!check_sig(cert, certs[indexes[0]]))
      return false;
   for(u32bit j = 0; j != indexes.size() - 1; j++)
      {
      if(!certs[indexes[j]].valid_time(current_time))
         return false;
      if(!check_sig(certs[indexes[j]], certs[indexes[j+1]]))
         return false;
      }
   return true;
   }

/*************************************************
* Find the parent of this certificate            *
*************************************************/
u32bit X509_Store::find_parent_of(const X509_Certificate& cert) const
   {
   for(u32bit j = 0; j != certs.size(); j++)
      {
      if(certs[j].subject == cert.issuer)
         {
         if(!compare_ids(certs[j].v2_subject_key_id, cert.v2_issuer_key_id))
            continue;
         if(!compare_ids(certs[j].v3_subject_key_id, cert.v3_issuer_key_id))
            continue;
         return j;
         }
      }
   return NO_CERT_FOUND;
   }

/*************************************************
* Check the CAs signature on a certificate       *
*************************************************/
bool X509_Store::check_sig(const X509_Certificate& cert,
                           const X509_Certificate& ca_cert) const
   {
   std::auto_ptr<X509_PublicKey> pub_key(ca_cert.subject_public_key());
   PK_Verifier* verifier = 0;

   try {
      std::vector<std::string> sig_info =
         split_on_slash(OIDS::lookup(cert.sig_algo.oid));

      if(sig_info.size() != 2 && sig_info.size() != 3)
         return false;
      if(sig_info[0] != pub_key->algo_name())
         return false;

      std::string padding = sig_info[1];
      Signature_Format format = IEEE_1363;

      if(sig_info.size() == 3)
         {
         if(sig_info[2] == "DER")
            format = DER_SEQUENCE;
         else
            throw Decoding_Error("Unknown encoding method " + sig_info[2]);
         }

      if(dynamic_cast<PK_Verifying_with_MR_Key*>(pub_key.get()))
         {
         PK_Verifying_with_MR_Key* sig_key =
            dynamic_cast<PK_Verifying_with_MR_Key*>(pub_key.get());
         verifier = get_pk_verifier(*sig_key, padding, format);
         }
      else if(dynamic_cast<PK_Verifying_wo_MR_Key*>(pub_key.get()))
         {
         PK_Verifying_wo_MR_Key* sig_key =
            dynamic_cast<PK_Verifying_wo_MR_Key*>(pub_key.get());
         verifier = get_pk_verifier(*sig_key, padding, format);
         }
      else
         return false;

      bool valid = verifier->verify_message(cert.tbs_certificate(),
                                            cert.signature());
      delete verifier;
      cert.trusted = valid;
      return valid;
   }
   catch(std::exception& e)
      {
      cert.trusted = false;
      delete verifier;
      }
   return false;
   }

/*************************************************
* Search for certificates in the store           *
*************************************************/
std::vector<X509_Certificate> X509_Store::find(const std::string& email,
                                               const std::string& name) const
   {
   std::vector<X509_Certificate> matching_certs;

   for(u32bit j = 0; j != certs.size(); j++)
      {
      const std::string common_name = certs[j].subject_info("Name");
      const std::string email_addr = certs[j].subject_info("RFC822");

      if((email == "" || email_addr.find(email) != std::string::npos) &&
         (name  == "" || common_name.find(name) != std::string::npos))
         matching_certs.push_back(certs[j]);
      }
   return matching_certs;
   }

/*************************************************
* Add a certificate to the store                 *
*************************************************/
void X509_Store::add_cert(const X509_Certificate& cert, bool trusted)
   {
   certs.push_back(cert);
   certs.back().trusted = trusted;
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::do_add_certs(DataSource& source, X509_Encoding encoding,
                              bool trusted)
   {
   while(!source.end_of_data())
      {
      try {
         X509_Certificate cert(source, encoding);
         add_cert(cert, trusted);
      }
      catch(X509_Certificate::X509_Error) {}
      catch(Decoding_Error) {}
      catch(Invalid_Argument) {}
      }
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::add_certs(DataSource& source, X509_Encoding encoding)
   {
   do_add_certs(source, encoding, false);
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::add_trusted_certs(DataSource& source, X509_Encoding encoding)
   {
   do_add_certs(source, encoding, true);
   }

/*************************************************
* PEM encode the set of certificates             *
*************************************************/
std::string X509_Store::PEM_encode() const
   {
   std::string cert_store;
   for(u32bit j = 0; j != certs.size(); j++)
      cert_store += certs[j].PEM_encode();
   return cert_store;
   }

}
